unit clGoogleCalendar;

interface

uses
  Classes, SysUtils, clWebDav, clHttpRequest, clXmlUtils, msxml, clTranslator, clUriUtils;
//https://developers.google.com/google-apps/calendar/v2/developers_guide_protocol?hl=ru#DeletingEvents
type
  TclGoogleCalendar = class(TComponent)
  private
    FWebDav: TclWebDav;
    FRequest: TclHttpRequest;
    FEventData: TStrings;

    FEventTitle: string;
    FEventID: string;
    FEventWhere: string;
    FEventEndTime: string;
    FEventContent: string;
    FAuthID: string;
    FEventStartTime: string;

    function GetEmail: string;
    function GetPassword: string;
    function GetUserAgent: string;
    procedure SetEmail(const Value: string);
    procedure SetPassword(const Value: string);
    procedure SetUserAgent(const Value: string);

    function NormalizeField(const AField: string): string;
    procedure InternalGetEventList(const AQuery: string; AList: TStrings);
  public
    constructor Create(Owner: TComponent); override;
    destructor Destroy; override;

    function Login: string;
    procedure Logout;

    procedure GetEventList(AList: TStrings); overload;
    procedure GetEventList(const AStartTime, AEndTime: string; AList: TStrings); overload;
    procedure GetEventList(const AText: string; AList: TStrings); overload;
    function CreateEvent: string;
    procedure RetrieveEvent(const AEventID: string);
    procedure UpdateEvent;
    procedure DeleteEvent(const AEventID: string);

    property AuthID: string read FAuthID;
    property EventID: string read FEventID;
    property WebDav: TclWebDav read FWebDav;
    property EventData: TStrings read FEventData;
  published
    property Email: string read GetEmail write SetEmail;
    property Password: string read GetPassword write SetPassword;
    property UserAgent: string read GetUserAgent write SetUserAgent;
    
    property EventTitle: string read FEventTitle write FEventTitle;
    property EventContent: string read FEventContent write FEventContent;
    property EventWhere: string read FEventWhere write FEventWhere;
    property EventStartTime: string read FEventStartTime write FEventStartTime;
    property EventEndTime: string read FEventEndTime write FEventEndTime;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Clever Internet Suite', [TclGoogleCalendar]);
end;

{ TclGoogleCalendar }

constructor TclGoogleCalendar.Create(Owner: TComponent);
begin
  inherited Create(Owner);

  FWebDav := TclWebDav.Create(nil);
  FRequest := TclHttpRequest.Create(nil);
  FWebDav.Request := FRequest;
  FEventData := TStringList.Create();
end;

function TclGoogleCalendar.CreateEvent: string;
var
  createRequest: string;
  Dom: IXMLDomDocument;
  node: IXMLDOMNode;
  title, content, location: string;
  ws: WideString;
begin
  FEventID := '';
  FEventData.Clear();

  ws := WideString(EventTitle);
  title := TclTranslator.TranslateToUtf8(ws);
  
  ws := WideString(EventContent);
  content := TclTranslator.TranslateToUtf8(ws);

  ws := WideString(EventWhere);
  location := TclTranslator.TranslateToUtf8(ws);

  createRequest := '<?xml version=''1.0'' encoding=''UTF-8''?>'#13#10 +
  '<entry xmlns=''http://www.w3.org/2005/Atom'''#13#10 +
    'xmlns:gd=''http://schemas.google.com/g/2005''>'#13#10 +
  '<category scheme=''http://schemas.google.com/g/2005#kind'''#13#10 +
    'term=''http://schemas.google.com/g/2005#event''></category>'#13#10 +
  '<title type=''text''>' + title + '</title>'#13#10 +
  '<content type=''text''>' + content + '</content>'#13#10 +
  '<gd:transparency'#13#10 +
    'value=''http://schemas.google.com/g/2005#event.opaque''>'#13#10 +
  '</gd:transparency>'#13#10 +
  '<gd:eventStatus'#13#10 +
    'value=''http://schemas.google.com/g/2005#event.confirmed''>'#13#10 +
  '</gd:eventStatus>'#13#10 +
  '<gd:where valueString=''' + location + '''></gd:where>'#13#10 +
  '<gd:when startTime=''' + EventStartTime + ''''#13#10 +
    'endTime=''' + EventEndTime + '''></gd:when>'#13#10 +
'</entry>';

  WebDav.Request.Clear();
  WebDav.Request.AddTextData(createRequest);
  WebDav.Request.Header.Authorization := 'GoogleLogin ' + AuthID;
  WebDav.Request.Header.ContentType := 'application/atom+xml';

  WebDav.Post('https://www.google.com/calendar/feeds/' + NormalizeField(Email) + '/private/full', WebDav.Response);

  Dom := CoDOMDocument.Create();
  Dom.loadXML(WideString(WebDav.Response.Text));

  node := Dom.selectSingleNode('*/link[@rel=''edit'']/@href');
  if (node <> nil) then
  begin
    FEventID := GetNodeText(node);
  end;

  Dom := nil;
  Result := FEventID;
end;

procedure TclGoogleCalendar.DeleteEvent(const AEventID: string);
begin
  if (AEventID = '') then
  begin
    raise Exception.Create('You must specify the Event ID');
  end;

  WebDav.Request.Clear();
  WebDav.Request.Header.Authorization := 'GoogleLogin ' + AuthID;
  WebDav.Request.Header.ContentType := 'application/atom+xml';
  WebDav.Request.Header.ExtraFields.Add('If-Match: *');

  WebDav.Delete(AEventID);
end;

destructor TclGoogleCalendar.Destroy;
begin
  FEventData.Free();
  FRequest.Free();
  FWebDav.Free();

  inherited Destroy();
end;

function TclGoogleCalendar.GetEmail: string;
begin
  Result := WebDav.UserName;
end;

procedure TclGoogleCalendar.GetEventList(const AStartTime, AEndTime: string; AList: TStrings);
begin
  InternalGetEventList('?start-min=' + AStartTime + '&start-max=' + AEndTime, AList);
end;

procedure TclGoogleCalendar.GetEventList(const AText: string; AList: TStrings);
begin
  InternalGetEventList('?q=' + GetEncodedUrl(AText), AList); //the title and content of an event
end;

procedure TclGoogleCalendar.GetEventList(AList: TStrings);
begin
  InternalGetEventList('', AList);
end;

function TclGoogleCalendar.GetPassword: string;
begin
  Result := WebDav.Password;
end;

function TclGoogleCalendar.GetUserAgent: string;
begin
  Result := FRequest.Header.UserAgent;
end;

procedure TclGoogleCalendar.InternalGetEventList(const AQuery: string; AList: TStrings);
var
  Dom: IXMLDomDocument;
  node: IXMLDOMNode;
  list: IXMLDOMNodeList;
begin
  AList.BeginUpdate();
  try
    WebDav.Request.Clear();
    WebDav.Request.Header.Authorization := 'GoogleLogin ' + AuthID;
    WebDav.Request.Header.ContentType := 'application/atom+xml';

    WebDav.Get('https://www.google.com/calendar/feeds/' + NormalizeField(Email) + '/private/full' + AQuery, AList);

    Dom := CoDOMDocument.Create();
    Dom.loadXML(WideString(AList.Text));

    AList.Clear();
    list := Dom.selectNodes('feed/entry');
    node := list.nextNode;
    while (node <> nil) do
    begin
      node := node.selectSingleNode('link[@rel=''edit'']/@href');
      if (node <> nil) then
      begin
        AList.Add(GetNodeText(node));
      end;
      node := list.nextNode;
    end;
  finally
    AList.EndUpdate();
  end;
end;

function TclGoogleCalendar.Login: string;
var
  ind: Integer;
begin
  if (AuthID <> '') then
  begin
    raise Exception.Create('The user already logged in');
  end;

  WebDav.Request.Clear();
  WebDav.Request.AddFormField('accountType', 'HOSTED_OR_GOOGLE');
  WebDav.Request.AddFormField('Email', Email);
  WebDav.Request.AddFormField('Passwd', Password);
  WebDav.Request.AddFormField('service', 'cl');
  WebDav.Request.AddFormField('source', UserAgent);

  WebDav.Post('https://www.google.com/accounts/ClientLogin', WebDav.Response);

  ind := Pos('Auth=', WebDav.Response.Text);

  if (ind > 0) then
  begin
    FAuthID := System.Copy(WebDav.Response.Text, ind, 10000);
  end;

  Result := FAuthID;
end;

procedure TclGoogleCalendar.Logout;
begin
  WebDav.Close();
  FAuthID := '';
  FEventID := '';
  FEventData.Clear();
end;

function TclGoogleCalendar.NormalizeField(const AField: string): string;
begin
  Result := StringReplace(AField, '@', '%40', [rfReplaceAll]);
end;

procedure TclGoogleCalendar.UpdateEvent;
var
  Dom: IXMLDomDocument;
  node: IXMLDOMNode;
  ws: WideString;
begin
  if (FEventID = '') or (FEventData.Count = 0) then
  begin
    raise Exception.Create('You must retrieve the Event before deleting');
  end;
  
  WebDav.Request.Clear();
  WebDav.Request.Header.Authorization := 'GoogleLogin ' + AuthID;
  WebDav.Request.Header.ContentType := 'application/atom+xml';

  Dom := CoDOMDocument.Create();
  Dom.loadXML(WideString(FEventData.Text));

  node := Dom.selectSingleNode('*/title');
  if (node <> nil) then
  begin
    ws := WideString(FEventTitle);
    node.text := WideString(TclTranslator.TranslateToUtf8(ws));
  end;

  node := Dom.selectSingleNode('*/content');
  if (node <> nil) then
  begin
    ws := WideString(FEventContent);
    node.text := WideString(TclTranslator.TranslateToUtf8(ws));
  end;

  node := Dom.selectSingleNode('*/gd:where');
  if (node <> nil) then
  begin
    ws := WideString(FEventWhere);
    SetAttributeValue(node, 'valueString', TclTranslator.TranslateToUtf8(ws));
  end;

  node := Dom.selectSingleNode('*/gd:when');
  if (node <> nil) then
  begin
    SetAttributeValue(node, 'startTime', FEventStartTime);
    SetAttributeValue(node, 'endTime', FEventEndTime);
  end;

  FEventData.Text := string(dom.xml);
  
  Dom := nil;
  
  WebDav.Put(FEventID, FEventData);
end;

procedure TclGoogleCalendar.RetrieveEvent(const AEventID: string);
var
  Dom: IXMLDomDocument;
  node: IXMLDOMNode;
  ws: WideString;
begin
  if (AEventID = '') then
  begin
    raise Exception.Create('You must specify the Event ID');
  end;

  FEventID := AEventID;
  FEventData.Clear();

  WebDav.Request.Clear();
  WebDav.Request.Header.Authorization := 'GoogleLogin ' + AuthID;
  WebDav.Request.Header.ContentType := 'application/atom+xml';
  
  WebDav.Get(FEventID, FEventData);

  Dom := CoDOMDocument.Create();
  Dom.loadXML(WideString(FEventData.Text));

  node := Dom.selectSingleNode('*/title');
  if (node <> nil) then
  begin
    ws := TclTranslator.TranslateFromUtf8(GetNodeText(node));
    FEventTitle := ws;
  end;

  node := Dom.selectSingleNode('*/content');
  if (node <> nil) then
  begin
    ws := TclTranslator.TranslateFromUtf8(GetNodeText(node));
    FEventContent := ws;
  end;

  node := Dom.selectSingleNode('*/gd:where');
  if (node <> nil) then
  begin
    ws := TclTranslator.TranslateFromUtf8(GetAttributeValue(node, 'valueString'));
    FEventWhere := ws;
  end;

  node := Dom.selectSingleNode('*/gd:when');
  if (node <> nil) then
  begin
    FEventStartTime := GetAttributeValue(node, 'startTime');
    FEventEndTime := GetAttributeValue(node, 'endTime');
  end;
  
  Dom := nil;
end;

procedure TclGoogleCalendar.SetEmail(const Value: string);
begin
  WebDav.UserName := Value;
end;

procedure TclGoogleCalendar.SetPassword(const Value: string);
begin
  WebDav.Password := Value;
end;

procedure TclGoogleCalendar.SetUserAgent(const Value: string);
begin
  FRequest.Header.UserAgent := Value;
end;

end.